/*
 * (C) 2006 - Grzegorz Milos - Cambridge University
 * (C) 2006 Nobuhiro Sakawa  Changes Date: Nov 2006
 * 
 *  $B>e5-Cx:n8"<T$O!$0J2<$N(B (1)$B!A(B(4) $B$N>r7o$+!$(BFree Software Foundation 
 *  $B$K$h$C$F8xI=$5$l$F$$$k(B GNU General Public License $B$N(B Version 2 $B$K5-(B
 *  $B=R$5$l$F$$$k>r7o$rK~$?$9>l9g$K8B$j!$K\%=%U%H%&%'%"!JK\%=%U%H%&%'%"(B
 *  $B$r2~JQ$7$?$b$N$r4^$`!%0J2<F1$8!K$r;HMQ!&J#@=!&2~JQ!&:FG[I[!J0J2<!$(B
 *  $BMxMQ$H8F$V!K$9$k$3$H$rL5=~$G5vBz$9$k!%(B
 *  (1) $BK\%=%U%H%&%'%"$r%=!<%9%3!<%I$N7A$GMxMQ$9$k>l9g$K$O!$>e5-$NCx:n(B
 *      $B8"I=<(!$$3$NMxMQ>r7o$*$h$S2<5-$NL5J]>Z5,Dj$,!$$=$N$^$^$N7A$G%=!<(B
 *      $B%9%3!<%ICf$K4^$^$l$F$$$k$3$H!%(B
 *  (2) $BK\%=%U%H%&%'%"$r!$%i%$%V%i%j7A<0$J$I!$B>$N%=%U%H%&%'%"3+H/$K;H(B
 *      $BMQ$G$-$k7A$G:FG[I[$9$k>l9g$K$O!$:FG[I[$KH<$&%I%-%e%a%s%H!JMxMQ(B
 *      $B<T%^%K%e%"%k$J$I!K$K!$>e5-$NCx:n8"I=<(!$$3$NMxMQ>r7o$*$h$S2<5-(B
 *      $B$NL5J]>Z5,Dj$r7G:\$9$k$3$H!%(B
 *  (3) $BK\%=%U%H%&%'%"$r!$5!4o$KAH$_9~$`$J$I!$B>$N%=%U%H%&%'%"3+H/$K;H(B
 *      $BMQ$G$-$J$$7A$G:FG[I[$9$k>l9g$K$O!$<!$N$$$:$l$+$N>r7o$rK~$?$9$3(B
 *      $B$H!%(B
 *    (a) $B:FG[I[$KH<$&%I%-%e%a%s%H!JMxMQ<T%^%K%e%"%k$J$I!K$K!$>e5-$NCx(B
 *        $B:n8"I=<(!$$3$NMxMQ>r7o$*$h$S2<5-$NL5J]>Z5,Dj$r7G:\$9$k$3$H!%(B
 *    (b) $B:FG[I[$N7ABV$r!$JL$KDj$a$kJ}K!$K$h$C$F!$(BTOPPERS$B%W%m%8%'%/%H$K(B
 *        $BJs9p$9$k$3$H!%(B
 *  (4) $BK\%=%U%H%&%'%"$NMxMQ$K$h$jD>@\E*$^$?$O4V@\E*$K@8$8$k$$$+$J$kB;(B
 *      $B32$+$i$b!$>e5-Cx:n8"<T$*$h$S(BTOPPERS$B%W%m%8%'%/%H$rLH@U$9$k$3$H!%(B
 * 
 *  $BK\%=%U%H%&%'%"$O!$L5J]>Z$GDs6!$5$l$F$$$k$b$N$G$"$k!%>e5-Cx:n8"<T$*(B
 *  $B$h$S(BTOPPERS$B%W%m%8%'%/%H$O!$K\%=%U%H%&%'%"$K4X$7$F!$$=$NE,MQ2DG=@-$b(B
 *  $B4^$a$F!$$$$+$J$kJ]>Z$b9T$o$J$$!%$^$?!$K\%=%U%H%&%'%"$NMxMQ$K$h$jD>(B
 *  $B@\E*$^$?$O4V@\E*$K@8$8$?$$$+$J$kB;32$K4X$7$F$b!$$=$N@UG$$rIi$o$J$$!%(B
 * 
 */

#include <xen_h/types.h>
#include <xen_h/wait.h>
#include <xen_h/mm.h>
#include <xen_h/hypervisor.h>
#include <xen_h/events.h>
#include <xen_h/os.h>
#include <xen_h/lib.h>
#include <xen_h/xenbus.h>
#include <xen_h/xen/io/console.h>

/* TODO - need to define BUG_ON for whole mini-os, need crash-dump as well */
#define BUG_ON(_cond)   do{if(_cond) do_exit();} while(0);

static inline struct xencons_interface *xencons_interface(void)
{
    return mfn_to_virt(start_info.console_mfn);
}

static inline void notify_daemon(void)
{
    /* Use evtchn: this is called early, before irq is set up. */
    notify_remote_via_evtchn(start_info.console_evtchn);
}

int xencons_send_no_notify(const char *data, unsigned len)
{	
    int sent = 0;
	struct xencons_interface *intf = xencons_interface();
	XENCONS_RING_IDX cons, prod;
	cons = intf->out_cons;
	prod = intf->out_prod;
	mb();
	BUG_ON((prod - cons) > sizeof(intf->out));

	while ((sent < len) && ((prod - cons) < sizeof(intf->out)))
		intf->out[MASK_XENCONS_IDX(prod++, intf->out)] = data[sent++];

	wmb();
	intf->out_prod = prod;
    
    return sent;
}

int xencons_send(const char *data, unsigned len)
{
    int sent;
    sent = xencons_send_no_notify(data, len);
	notify_daemon();

	return sent;
}


#include <t_services.h>
#include <serial.h>
#include <hw_serial.h>
	
#define	SERIAL_BUFSZ	256		/* $B%I%i%$%P$N%P%C%U%!%5%$%:(B */
typedef struct serial_port_initialization_block {
	ID	rcv_semid;	/* $B<u?.%P%C%U%!4IM}MQ%;%^%U%)$N(BID */
	ID	snd_semid;	/* $BAw?.%P%C%U%!4IM}MQ%;%^%U%)$N(BID */
} SPINIB;
typedef struct serial_port_control_block {
	const SPINIB *spinib;	/* $B%7%j%"%k%]!<%H=i4|2=%V%m%C%/(B */
	SIOPCB	*siopcb;	/* $B%7%j%"%k(BI/O$B%]!<%H4IM}%V%m%C%/(B */
	BOOL	openflag;	/* $B%*!<%W%s:Q$_%U%i%0(B */
	UINT	ioctl;		/* $BF0:n@)8f$N@_DjCM(B */

	UINT	rcv_read_ptr;	/* $B<u?.%P%C%U%!FI=P$7%]%$%s%?(B */
	UINT	rcv_write_ptr;	/* $B<u?.%P%C%U%!=q9~$_%]%$%s%?(B */
	UINT	rcv_count;	/* $B<u?.%P%C%U%!Cf$NJ8;z?t(B */
	char	rcv_fc_chr;	/* $BAw$k$Y$-(B START/STOP */
	BOOL	rcv_stopped;	/* STOP $B$rAw$C$?>uBV$+!)(B */

	UINT	snd_read_ptr;	/* $BAw?.%P%C%U%!FI=P$7%]%$%s%?(B */
	UINT	snd_write_ptr;	/* $BAw?.%P%C%U%!=q9~$_%]%$%s%?(B */
	UINT	snd_count;	/* $BAw?.%P%C%U%!Cf$NJ8;z?t(B */
	BOOL	snd_stopped;	/* STOP $B$r<u$1<h$C$?>uBV$+!)(B */

	char	rcv_buffer[SERIAL_BUFSZ];	/* $B<u?.%P%C%U%!(B */
	char	snd_buffer[SERIAL_BUFSZ];	/* $BAw?.%P%C%U%!(B */
} SPCB;

extern SPCB	spcb_table[1];
/*
 *  $B%]%$%s%?$N%$%s%/%j%a%s%H(B
 */
#define INC_PTR(ptr)		{ if (++ptr == SERIAL_BUFSZ) ptr = 0; }

static void handle_input(int port, struct pt_regs *regs)
{
	struct xencons_interface *intf = xencons_interface();
	XENCONS_RING_IDX cons, prod;
	char 	*buf;
	SPCB	*spcb;
	char	c;

	cons = intf->in_cons;
	prod = intf->in_prod;
	mb();
	BUG_ON((prod - cons) > sizeof(intf->in));
	spcb = spcb_table; //(SPCB *) exinf;
	while (cons != prod) {
		buf = (char *)intf->in+MASK_XENCONS_IDX(cons,intf->in);
        	buf[1] = '\0';
		cons++;
		/*
		 *  $B<u?.$7$?J8;z$r<u?.%P%C%U%!$KF~$l$k!%(B
		 */
		spcb->rcv_buffer[spcb->rcv_write_ptr] = buf[0];
		INC_PTR(spcb->rcv_write_ptr);
		if (buf[0] =='\r')
			printk("\n");
        	printk("%s", buf);
		if (spcb->rcv_count == 0) {
			isig_sem(spcb->spinib->rcv_semid);
		}
		spcb->rcv_count++;

	}

	mb();
	intf->in_cons = cons;

	notify_daemon();

}

int xencons_init(void)
{
	int err;

	if (!start_info.console_evtchn)
		return 0;

	err = bind_evtchn(start_info.console_evtchn, handle_input);
	if (err <= 0) {
		printk("XEN console request chn bind failed %i\n", err);
		return err;
	}

	/* In case we have in-flight data after save/restore... */
	notify_daemon();

	return 0;
}

void xencons_resume(void)
{
	(void)xencons_init();
}

